/*****************************************************************************
 *
 *  Parameter.cc                               XMM Science Analysis Subsystem
 *
 *  Parameter of a task
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: Parameter.cc,v 1.23 2002/12/17 17:53:49 jbakker Exp $
 *
 *****************************************************************************/

#include "Parameter.h"         // Application headers
#include "ChoiceChildParameters.h"
#include "StringToken.h"
#include "QuotedString.h" // utils
#include <Evaluator.h> // selectlib

#include <iostream>          // C++ headers
#include <errstr.h>
#include <vector>
#include <string>
#include <list>


// BEGIN private classes
class Parameter::ParameterViews : public std::list<ParameterView* > {};
// END private classes



const std::string          Parameter::_emptyString;
const std::vector<int>     Parameter::_emptyBoolList;  // Temporary work-around for egcs
const std::vector<int>     Parameter::_emptyIntList;
const std::vector<double>  Parameter::_emptyDoubleList;
const std::vector<std::string>  Parameter::_emptyStringList;


Parameter::Parameter(const std::string& name, const std::string& prompt) :
  _name(name),
  _prompt(prompt), 
  _isSet(false),
  _isMandatory(false),
  _constraints("true"),
  _views(new ParameterViews),
  _children(0)
{
  _children=new ChoiceChildParameters(this);
}


Parameter::Parameter(const Parameter& p) :
  _name(p._name),
  _prompt(p._prompt), 
  _isSet(p._isSet),
  _isMandatory(p._isMandatory),
  _constraints(p._constraints),
  _views(new ParameterViews),
  _children(0)
{
  _children=new ChoiceChildParameters(this);
}


Parameter::~Parameter()
{
  for (ParameterViews::iterator it=_views->begin();it!=_views->end();it++)
    if ( (*it)->isSticky() ) delete (*it); // 

  delete _views;
  delete _children;
}


// better wording would have been: connectView
void Parameter::addView(ParameterView* view) // NEW interface
{
  if (view) _views->push_back(view);  
}

// better wording would have been: disconnectView
void Parameter::deleteView(ParameterView* view) 
{
  if (!view) return;
  
  _views->remove(view);
  if (view->isSticky()) delete view; // todo: is 'view' element of _views?
}

void Parameter::notifyView()
{
  ParameterViews::const_iterator it;
  for (it=_views->begin();it!=_views->end();it++) (*it)->updateView();
}

void Parameter::wrongType() const
{
  errstr << "Parameter '" << _name << "' is used with wrong type" 
         << Sas::error("ParamWrongType");
}

bool Parameter::boolValue() const
{
    wrongType();
    return false;
}

int Parameter::intValue() const
{
    wrongType();
    return 0;
}

double Parameter::realValue() const
{
    wrongType();
    return 0;
}

const std::string& Parameter::stringValue() const
{
    wrongType();
    return _emptyString;    // Cannot return reference to temporary
}

std::vector<int> Parameter::boolListValue() const
{
    wrongType();
    return _emptyBoolList;
}

std::vector<int> Parameter::intListValue() const
{
    wrongType();
    return _emptyIntList;
}

std::vector<double> Parameter::realListValue() const
{
    wrongType();
    return _emptyDoubleList;
}

std::vector<std::string> Parameter::stringListValue() const
{
    wrongType();
    return _emptyStringList;
}

std::ostream& operator<<(std::ostream& os, const Parameter& p)
{
  QuotedString s(p.asString());
  s.quotes('\'','\'');
  return os << s;
}


/*** Write help about parameter type, defaults, etc ***/

void Parameter::writeHelpToStream(std::ostream& os) const
{
  // name -- mandatory,type=list<int>,constraints="",default=""

  os << "  " << _name << " -- "
     << (_isMandatory?"mandatory":"optional")
     << ",type=";
  if (size() < 0 ) os << getTypeName();
  else os << "list<" << getTypeName() << '>';

  if (_constraints != "true") {
    os <<",constraints=";
    writeConstraint(os);
  }
  
  QuotedString s=defaultAsString();
  if (!s.empty()) {
    s.quotes('\'','\'');
    os << ",default=" << s;
  }
  
  if(!_prompt.empty()) os << "; " << _prompt;
  os << std::endl;
  
  //    writeConstraint(os);
}

void Parameter::writeConstraint(std::ostream& os) const
{
  os << QuotedString(_constraints);
}


bool Parameter::setValue(const Parameter* p)
{
  if (this == p) return false;
  if (getType() != p->getType() || getType() == Undefined) return false;

  return assignValueOf(*p);
}


void Parameter::addChild(const std::string& choice,Parameter* child)
{
  _children->add(choice,child);
}

ChoiceChildParameters&  Parameter::children() const
{
  return *_children;
}

std::list<std::string > Parameter::choices()
{
 return _children->choices();
}


void Parameter::setConstraints(const std::string& expr)
{
  if (Evaluator::isValidExpression(expr)) {

    Evaluator evaluator( expr );
    {
      NumericSymbols ids = evaluator.numericIdentifiers();
      for (unsigned int i=0;i<ids.name.size();i++) {
	if (_name != ids.name[i])
	  errstr << "Parameter("<<_name<<") has "
		 << "invalid numeric variable \"" << ids.name[i]
		 << "\" in expression: \"" << expr << "\"." 
		 << Sas::error("ParamInvalidConstraints");
      }
    }
    {
      StringSymbols ids = evaluator.stringIdentifiers();
      for (unsigned int i=0;i<ids.name.size();i++) {
	if (_name != ids.name[i])
	  errstr << "Parameter("<<_name<<") has "
		 << "invalid std::string variable \"" << ids.name[i]
		 << "\" in expression: \"" << expr << "\"." 
		 << Sas::error("ParamInvalidConstraints");
      }
    }
    _constraints=expr;

  }
  else errstr << "Parameter("<<_name<<") has "
	      << "invalid constraints expression '" << expr << "'." 
	      << Sas::error("ParamInvalidConstraints");
}

const std::string& Parameter::constraints() const
{
  return _constraints;
}
