/*****************************************************************************
 *
 *  RealListParameter.cc                        XMM Science Analysis Subsystem
 *
 *  Real std::list task parameter
 *
 *  (c) 1998  ESA  -  Written by Jon Brumfitt, Aurora
 *
 *  $Id: RealListParameter.cc,v 1.23 2002/12/17 17:53:49 jbakker Exp $
 *
 *****************************************************************************/

#include <sstream>             // C++ headers
#include <iostream>
#include <string>
#include <vector>

#include "RealListParameter.h"     // Application headers
#include "StringToken.h"
#include <errstr.h>
#include <StringFns.h>
#include <Evaluator.h>
#include "QuotedString.h" // utils


//================================================================
//  Class RealListParameter
//================================================================

/*** Format the parameter value as a std::string in quotes ***/

std::string RealListParameter::asString() const
{
    std::ostringstream ostr;
    std::vector<double>::const_iterator it = _value.begin();
    if(it != _value.end())
        ostr << QuotedString(doubleAsString(*it++));
    while(it != _value.end())
        ostr << ' ' << QuotedString(doubleAsString(*it++));
    return ostr.str();
}


/*** Format the parameter value as a std::string in quotes ***/

std::string RealListParameter::defaultAsString() const
{
    std::ostringstream ostr;
    std::vector<double>::const_iterator it = _default.begin();
    if(it != _default.end())
        ostr << QuotedString(doubleAsString(*it++));
    while(it != _default.end())
        ostr << ' ' << QuotedString(doubleAsString(*it++));
    return  ostr.str();
}


/*** Format a single std::list element as a std::string ***/

std::string RealListParameter::doubleAsString(double value) const
{
    std::ostringstream ostr;
    ostr.precision(12);
    ostr << value;
    return ostr.str();
}


/***  Parse std::string representation and assign to 'value' if it is valid ***/

bool RealListParameter::parse(const std::string& str, std::vector<double>& value)
{
// TEMPORARY WORKAROUND TO ALLOW TIMES TO BE QUOTED

    std::list<std::string> stringTokens = StringFns::splitStringList(std::string(str));
//    list<string> stringTokens = StringFns::splitString(string(str)," \t");

    bool ok = true;
    bool first = true;
    std::list<std::string>::const_iterator it = stringTokens.begin();
    while(ok && it != stringTokens.end())
    {
        std::string token = *it++;
	double item = 0.0;
	ok &= parseDouble(token, item, first);
	first = false;
	if(ok)
	    value.push_back(item);
	else {
	  errstr << "Parameter '" << _name << "'" << Sas::error("ParamWrongType");
	  ok = false;
	}
    }
    return ok;
}


/*** Parse a single std::list element ***/

bool RealListParameter::parseDouble(const std::string& s, double& result, bool)
{
    StringToken token(s);
    result = token.asDouble();

    return token.isValid();
}




/*** Revert to default value ***/

void RealListParameter::revertToDefault()
{
    _value = _default;
    notifyView();
}


/*** Test whether value is allowable for the parameter ***/

bool RealListParameter::isValid(const std::vector<double>& value) const
{
  Evaluator evaluator( constraints() );
  NumericSymbols ids = evaluator.numericIdentifiers();  
  if (ids.valuePtr.size() == 0) return evaluator.evalLogic();
      
  std::vector<double>::const_iterator it = value.begin();
  while(it != value.end()) {
    for (unsigned int i=0;i<ids.valuePtr.size();i++) *ids.valuePtr[i]=*it;
    if (!evaluator.evalLogic()) return false;
    it++;
  }
  return true;
}


/*** Set the parameter to the specified value ***/

bool RealListParameter::setValue(const std::vector<double>& value)
{
    bool valid = isValid(value);
    if(valid)
    {
        _value = value;
	setDefined();
	notifyView();
    }
    return valid;
}


/*** Set the default to the specified value ***/

bool RealListParameter::setDefault(const std::vector<double>& value)
{
    bool valid = isValid(value);
    if(valid)
        _default = value;
    return valid;
}


/*** Set the value, specified as a std::string ***/

bool RealListParameter::setNewValue(const std::string& str)
{
    std::vector<double> values;
    bool ok =  parse(str, values) && setValue(values);

    if(!ok)
       reportRangeError(str);

    return true;
}


/***  Set the default, specified as a std::string ***/

bool RealListParameter::setNewDefault(const std::string& str)
{
    std::vector<double> values;
    bool ok = parse(str, values) && setDefault(values);

    if(!ok)
        reportRangeError(str);

    return true;
}


int RealListParameter::size() const
{
    return _value.size();
}


/*** Report that value is out of range ***/

void RealListParameter::reportRangeError(const std::string& valueString)
{
  errstr << "Parameter '" << _name << "';" << std::endl
	 << "value '" << valueString << "' is outside range ";
  writeConstraint(errstr);
  errstr << Sas::error("ParamNotInRange");
}
